#include "raine.h"
#include "debug.h"

static UBYTE *MemoryPool[256];	// Pointers to allocated memory areas
static int MemoryPoolCount;	// Number of items in memory pool
static int MemoryPoolSize;	// Size of all items in memory pool

// AllocateMem():
// Allocates a space memory, size bytes long
// Returns a pointer on success, or NULL on failiure. Also generates
// a raine error message on failure...

UBYTE *AllocateMem(ULONG size)
{
   UBYTE *memptr;

   // Attempt to allocate memory
   // --------------------------

   memptr = (UBYTE*) malloc(size);

   if(!(memptr)){

      // Raine error (eg. for gui)
      // -------------------------

      sprintf(load_debug+strlen(load_debug),"ERROR: Unable to allocate memory\n");
      sprintf(load_debug+strlen(load_debug),"\n");
      sprintf(load_debug+strlen(load_debug),"Failed to obtain 0x%08X bytes\n",size);
      sprintf(load_debug+strlen(load_debug),"\n");
      sprintf(load_debug+strlen(load_debug),"To fix this, buy more ram, or increase the available dpmi memory to 65535kb. ");
      sprintf(load_debug+strlen(load_debug),"Read rainefaq.txt for help with this (don't email us about it, thanks).\n");
      sprintf(load_debug+strlen(load_debug),"\n");

      load_error |= LOAD_FATAL_ERROR;

      #ifdef HACKS
      print_debug("ERROR: Unable to allocate 0x%08X bytes of memory\n",size);
      #endif

      return NULL;

   }
   else{

      // Check memory pool is not full up (unlikely)
      // -------------------------------------------

      if(MemoryPoolCount==256){
      #ifdef HACKS
      print_debug("ERROR: Too many banks in memory pool (256)\n");
      #endif
      }

      // Add to memory pool list
      // -----------------------

      MemoryPool[MemoryPoolCount++]=memptr;
      MemoryPoolSize+=size;
      #ifdef HACKS
      print_debug("Allocated 0x%08X bytes of memory\nMemory Blocks: 0x%02X\nAllocated Memory Total: 0x%08X\n",size,MemoryPoolCount,MemoryPoolSize);
      //print_debug("Now memory free: 0x%08x\n", malloc(0));
      #endif
   }

   return memptr;
}

// FreeMem():
// Deallocates a specific memory resource, memptr

void FreeMem(UBYTE *memptr)
{
   int ta;

   if(memptr!=NULL){

   free(memptr);

   for(ta=0;ta<MemoryPoolCount;ta++){
      if(MemoryPool[ta]==memptr) MemoryPool[ta]=NULL;
   }

   }
}

// ResetMemoryPool():
// Call this to start/reset the ingame memory allocation list. Should
// be called only in one place, before loadgame.

void ResetMemoryPool(void)
{
   int ta;

   for(ta=0;ta<256;ta++){
      MemoryPool[ta]=NULL;
   }

   MemoryPoolCount=0;
   MemoryPoolSize=0;
}

void FreeMemoryPool(void)
{
   int ta;

   #ifdef HACKS
   print_debug("START: FreeMemoryPool();\n");
   //print_debug("Before memory free: 0x%08x\n", malloc(0));
   #endif

   for(ta=0;ta<MemoryPoolCount;ta++){
      if(MemoryPool[ta]!=NULL) FreeMem(MemoryPool[ta]);
   }

   ResetMemoryPool();

   #ifdef HACKS
   //print_debug("After memory free: 0x%08x\n", malloc(0));
   print_debug("END: FreeMemoryPool();\n");
   #endif
}

#ifdef MEMORY_DEBUG

#undef malloc
#undef realloc
#undef free

struct sMemEntry
{
   void * entry;
   char * filename;
   ULONG  line;

   struct sMemEntry * next;
};

static struct sMemEntry * list = NULL;

// Init the memory list

void mbInitPurify(void)
{
   if (list)
      DonePurify();

   list = NULL;
}

// Clear the list, reporting leaked memory

void mbDonePurify(void)
{
   struct sMemEntry * node = list;

   while (node)
   {
      struct sMemEntry * tempnode = node;

      #ifdef HACKS
      print_debug("Memory Leak in <%s> at line %d  ->  %08xh\n", node->filename, node->line, node->entry);
      #endif
      free(node->entry);	// Don't free, incase it is still in use

      node = node->next;
      free(tempnode);
   }

   list = NULL;
}

// malloc replacement

void *mymalloc(ULONG size, char * fname, ULONG fline)
{
   struct sMemEntry * node;

   if (!size){			// Malloc(0) is used to get actual Heap Pointer
      #ifdef HACKS
      print_debug("mymalloc: malloc(0) in file <%s> at line %d.\n", fname, fline);
      #endif
      return malloc(size); 		// thus is not a memory leak
   }

   node = malloc(sizeof(struct sMemEntry));

   node->next = list;
   node->filename = fname;
   node->line = fline;
   node->entry = malloc(size);

   list = node;

   return node->entry;
}

// realloc replacement

void *myrealloc(void *ptr, ULONG size, char *fname, ULONG fline)
{
   struct sMemEntry *node;
   struct sMemEntry *prev;
   struct sMemEntry *next;

   if(ptr!=NULL){

   prev = NULL;
   node = list;

   while(node){

      next = node->next;

      if(node->entry == ptr){
         if(prev != NULL){
            prev->next = next;
         }
         else{
            list = next;
         }
         free(node);
      }
      else{
         prev = node;
      }
      node = next;
   }

   }

   // -----

   node = malloc(sizeof(struct sMemEntry));

   node->next = list;
   node->filename = fname;
   node->line = fline;
   node->entry = realloc(ptr, size);

   list = node;

   return node->entry;
}

// free replacement

void myfree(void * ptr, char * fname, ULONG fline)
{
   int found;
   struct sMemEntry *node;
   struct sMemEntry *prev;
   struct sMemEntry *next;

   found = 0;

   prev = NULL;
   node = list;

   while(node){

      next = node->next;

      if(node->entry == ptr){
         if(prev != NULL){
            prev->next = next;
         }
         else{
            list = next;
         }
         free(node->entry);
         free(node);
         found++;
      }
      else{
         prev = node;
      }
      node = next;
   }

   #ifdef HACKS
   if(!found){
      print_debug("myfree: Trying to delete unexistant pointer $%08x in file <%s> at line %d.\n", ptr, fname, fline);
   }
   if(found > 1){
      print_debug("myfree: Pointer was found %d times $%08x in file <%s> at line %d.\n", found, ptr, fname, fline);
   }
   #endif
}

#endif