/* esp.c
 */

#include "esp.h"
#include <string.h>

/* Serial I/O Registers */

#define SER0R	0xea
#define SER0L	0xeb
#define SER1R	0xec
#define SER1L	0xed
#define SER2R	0xee
#define SER2L	0xef
#define SER3R	0xf0
#define SER3L	0xf1

/* Multiplier Registers */

#define MACL	0xf2
#define MACH	0xf3

/* Address Generator Registers */

#define DIL	0xf4	/* read */
#define MEMSIZ	0xf4	/* write */
#define DLENGTH	0xf5
#define ABASE	0xf6
#define BBASE	0xf7
#define DBASE 	0xf8

/* Control Registers */

#define SIGREG	0xf9
#define CCR	0xfa
#define CMR	0xfb

/* Constant Memory */

#define MINUS_ONE		0xfc
#define NEGATIVE_FULL_SCALE	0xfd
#define POSITIVE_FULL_SCALE	0xfe
#define ZERO			0xff

/* Condition Code Register Bits */

#define CONDITION_CODE_N	(1 << 23)
#define CONDITION_CODE_C	(1 << 22)
#define CONDITION_CODE_V	(1 << 21)
#define CONDITION_CODE_LT	(1 << 20)
#define CONDITION_CODE_Z	(1 << 19)
#define CONDITION_CODE_NOT	(1 << 18)

/* Host Read/Write Addresses */

#define REGISTER_23_16  0x00
#define REGISTER_15_8   0x01
#define REGISTER_7_0    0x02

#define INSTRUCTION_47_40       0x03
#define INSTRUCTION_39_32       0x04
#define INSTRUCTION_31_24       0x05
#define INSTRUCTION_23_16       0x06
#define INSTRUCTION_15_8        0x07
#define INSTRUCTION_7_3         0x08

#define DIL_23_16       0x09
#define DIL_15_8        0x0a
#define DIL_7_0         0x0b    /* always read as zero */

#define DOL_23_16       0x0c
#define DOL_15_8        0x0d
#define DOL_7_0         0x0e    /* always read as 0xff */

#define DADR_23_16      0x0f
#define DADR_15_8       0x10
#define DADR_7_4        0x11

#define HOST_CONTROL_2_0        0x12
#define RAM_ACCESS_CONTROL_7_6  0x14

#define PROGRAM_COUNTER                 0x16
#define INTERNAL_REFRESH_COUNTER        0x17
#define HOST_SERIAL_CONTROL             0x18

#define HALT_ENABLE     0x1f    /* write only */
#define FRAME_COUNTER   0x1f    /* read only */

#define READ_SELECT_REGISTER_INSTRUCTION        0x80
#define WRITE_SELECT_REGISTER                   0xa0
#define WRITE_SELECT_INSTRUCTION                0xc0
#define WRITE_SELECT_REGISTER_INSTRUCTION       0xe0

enum ALU_OPCODES {
	ALU_ADD,
	ALU_SUB,
	ALU_ADDU,
        ALU_SUBU,
	ALU_CMP,
	ALU_AND,
	ALU_OR,
	ALU_XOR,
	ALU_ABS,
	ALU_MOV,
	ALU_ASL2,
	ALU_ASL8,
	ALU_LS15,
	ALU_DIFF,
	ALU_ASR,
	ALU_END
};

void InitializeEnsoniqSignalProcessor (ESP *esp)

{
        memset(esp, 0, sizeof(ESP));
        return;
}

void ShutdownEnsoniqSignalProcessor (ESP *esp)

{
        /* Nothing, but for the sake of "orthogonality", "reflexivity", maybe
         * "duality", or whatever, of functions.
         */

        return;
}

int ReadEnsoniqSignalProcessor (ESP *esp, int address)

{
        int     result;

        switch (address) {

                case REGISTER_23_16 :   result = esp->register_23_16; break;
                case REGISTER_15_8 :    result = esp->register_15_8; break;
                case REGISTER_7_0 :     result = esp->register_7_0; break;

                case INSTRUCTION_47_40 :
                        result = esp->instruction_47_40;
                        break;
                case INSTRUCTION_39_32 :
                        result = esp->instruction_39_32;
                        break;
                case INSTRUCTION_31_24 :
                        result = esp->instruction_31_24;
                        break;
                case INSTRUCTION_23_16 :
                        result = esp->instruction_23_16;
                        break;
                case INSTRUCTION_15_8 :
                        result = esp->instruction_15_8;
                        break;
                case INSTRUCTION_7_3 :
                        result = esp->instruction_7_3;
                        break;

                case DIL_23_16 :        result = esp->dil_23_16; break;
                case DIL_15_8 :         result = esp->dil_15_8; break;
                case DIL_7_0 :          result = 0x00; break;

                case DOL_23_16 :        result = esp->dol_23_16; break;
                case DOL_15_8 :         result = esp->dol_15_8; break;
                case DOL_7_0 :          result = 0xff; break;

                case DADR_23_16 :       result = esp->dadr_23_16; break;
                case DADR_15_8 :        result = esp->dadr_15_8; break;
                case DADR_7_4 :         result = esp->dadr_7_4; break;

                case HOST_CONTROL_2_0 :
                        result = esp->host_control;
                        break;
                case RAM_ACCESS_CONTROL_7_6 :
                        result = esp->ram_access_control;
                        break;

                case PROGRAM_COUNTER :
                        result = esp->program_counter;
                        break;
                case INTERNAL_REFRESH_COUNTER :
                        result = esp->internal_refresh_counter;
                        break;
                case HOST_SERIAL_CONTROL :
                        result = esp->host_serial_control;
                        break;

                case FRAME_COUNTER :    result = -1; break;     /* not supported */

                case READ_SELECT_REGISTER_INSTRUCTION :
                        result = esp->read_select_register_instruction;
                        break;
                case WRITE_SELECT_REGISTER :
                        result = esp->write_select_register;
                        break;
                case WRITE_SELECT_INSTRUCTION :
                        result = esp->write_select_instruction;
                        break;
                case WRITE_SELECT_REGISTER_INSTRUCTION :
                        result = esp->write_select_register_instruction;
                        break;

                default:        result = -1; break;      /* problem! */
        }

        return result;
}

void WriteEnsoniqSignalProcessor (ESP *esp, int address, int data)

{
        int     where;

        switch (address) {

                case REGISTER_23_16 :
                        esp->register_23_16 = data;
                        t = esp->registers[where];
                        t &= 0xffff;
                        t += data << 16;
                        esp->registers[where] = t;
                        break;
                case REGISTER_15_8 :
                        esp->register_15_8 = data;
                        t = esp->registers[where];
                        t &= 0xff00ff;
                        t += data << 8;
                        esp->registers[where] = t;
                        break;
                case REGISTER_7_0 :
                        esp->register_7_0 = data;
                        t = esp->registers[where];
                        t &= 0xffff00;
                        t += data;
                        esp->registers[where] = t;
                        break;

                case INSTRUCTION_47_40 :
                        esp->instruction_47_40 = data;
                        t = esp->instructions[where];
                        t &= 0xffff00;
                        t += data;
                        esp->registers[where] = t;
                        break;
                case INSTRUCTION_39_32 :
                        result = esp->instruction_39_32;
                        break;
                case INSTRUCTION_31_24 :
                        result = esp->instruction_31_24;
                        break;
                case INSTRUCTION_23_16 :
                        result = esp->instruction_23_16;
                        break;
                case INSTRUCTION_15_8 :
                        result = esp->instruction_15_8;
                        break;
                case INSTRUCTION_7_3 :
                        result = esp->instruction_7_3;
                        break;

                case DIL_23_16 :        result = esp->dil_23_16; break;
                case DIL_15_8 :         result = esp->dil_15_8; break;
                case DIL_7_0 :          result = 0x00; break;

                case DOL_23_16 :        result = esp->dol_23_16; break;
                case DOL_15_8 :         result = esp->dol_15_8; break;
                case DOL_7_0 :          result = 0xff; break;

                case DADR_23_16 :       result = esp->dadr_23_16; break;
                case DADR_15_8 :        result = esp->dadr_15_8; break;
                case DADR_7_4 :         result = esp->dadr_7_4; break;

                case HOST_CONTROL_2_0 :
                        result = esp->host_control;
                        break;
                case RAM_ACCESS_CONTROL_7_6 :
                        result = esp->ram_access_control;
                        break;

                case PROGRAM_COUNTER :
                        result = esp->program_counter;
                        break;
                case INTERNAL_REFRESH_COUNTER :
                        result = esp->internal_refresh_counter;
                        break;
                case HOST_SERIAL_CONTROL :
                        result = esp->host_serial_control;
                        break;

                case FRAME_COUNTER :    result = -1; break;     /* not supported */

                case READ_SELECT_REGISTER_INSTRUCTION :
                        result = esp->read_select_register_instruction;
                        break;
                case WRITE_SELECT_REGISTER :
                        result = esp->write_select_register;
                        break;
                case WRITE_SELECT_INSTRUCTION :
                        result = esp->write_select_instruction;
                        break;
                case WRITE_SELECT_REGISTER_INSTRUCTION :
                        result = esp->write_select_register_instruction;
                        break;

                default:        result = -1; break;      /* problem! */
        }

}

int ExecuteInstruction (ESP *esp)
/* Return true if ended. */

{
        esp->program_counter = 0;
        return 0;
}

void RunEnsoniqSignalProcessor (ESP *esp)

{
        esp->program_counter = 0x00;
        while (!ExecuteInstruction(esp))
              ;
        return;
}
