/* Yamaha FM/OP-N emulator v1.01
 *
 * Copyright (c) 1999 Richard Mitton
 * All unauthorised distribution prohibited.
 *
 * Supports:
 *   YM2203 (OPN)
 *   YM2610 (OPNA?)
 *   YM2612 (OPNB?)
 *
 * Features not done yet:
 *   3-slot mode
 *   CSW mode
 *   Detune
 *   Pre-scalers
 *   Timers
 *   LFO
 *   SSG
 *   DAC
 *   ADPCM
 *
 * Some connection methods don't work too well.
 * The KEY-ON system isn't 100% correct, but should generally work.
 *
 * Changes:
 *   v1.01: Various fixes, YM2610/YM2612 support, debug handler.
 *   v1.00: Initial release.
 */

#include <stdio.h>
//#include <dos.h>
#ifdef RAINE_DOS
//#include <pc.h>
#include <asm/io.h>
#endif
#include "opn.h"

//#ifndef RAINE_DOS

#define outportb(a,b)	outb(b,a)
#define inportb		inb

//#endif

/* Local storage */
static int num_OPN, OPN_type, OPN_initialised=FALSE;
static OPN YM[2];
static byte OPL3_reg[0x200];
static int OPL3_base;
static void (*msg_handler)(char *text);

/* Local functions */
static void OPN_WriteReg(OPN *ym, byte reg, byte value);

/*
 * Debug aid -
 * #define ONLY_PLAY_CHANNEL to just play the one channel.
 * define it as the channel number (1->3 for 2203, 1->6 for other)
 */
/* #define ONLY_PLAY_CHANNEL 4 */

/*
 * If the user has set a debug message handler, this will call it with
 * a message.
 */
static DEF_INLINE void DebugMessage(char *text)
{
   if (msg_handler) msg_handler(text);
}

/*
 * This function writes a value to the OPL3.
 */
void WriteOPL3Reg(int reg, byte value)
{
   int i, base;

   OPL3_reg[reg] = value;   
   base = OPL3_base;
   if (reg & 0x100)
   {
      reg &= 0xff;
      base += 2;
   }

   /* Wait before writing to the port.
    * The most accurate way to do this is apparently to read the
    * port 35 times.
    */
   i = 35;
   do {
      inportb(base);
   } while(--i);
   outportb(base, reg);

   /* Wait again before writing to the port.
    * Only read 7 times for the data write.
    */
   i = 7;
   do {
      inportb(base);
   } while(--i);
   outportb(base+1, value);
}

/*
 * This function reads a (cached) OPL3 register.
 */
DEF_INLINE byte ReadOPL3Reg(int reg)
{
   return OPL3_reg[reg];
}

/*
 * This function reads the OPL3 status register.
 */
byte ReadOPL3Status(void)
{
   int i;

   /* Wait before reading from the port.
    * The most accurate way to do this is apparently to read the
    * port 35 times.
    */
   i = 35;
   do {
      inportb(OPL3_base);
   } while(--i);
   return inportb(OPL3_base);
}

/*
 * Reset the OPL3 chip.
 */
void ResetOPL3(void)
{
   int r;

   /* zero all registers: 1 to F5 (inclusive) */
   for (r=1;r<=0xf5;r++)
   {
      WriteOPL3Reg(r, 0x00);
      WriteOPL3Reg(r+0x100, 0x00);
   }

   /* set OPL3 mode */
   WriteOPL3Reg(0x105, 0x01);

   /* set 4-OP mode */
   WriteOPL3Reg(0x104, 0x3f);
}

/*
 * The OPL2 detection method is the official Adlib method.
 * The OPL3 method isn't. :-(
 */
int DetectOPL3(void)
{
#if 0				// crashes Antiriad's PC!!!

	byte status1, status2;

/* Detect OPL2 first:
 |      1)  Reset both timers by writing 60h to register 4.
 |      2)  Enable the interrupts by writing 80h to register 4.  NOTE: this
 |          must be a separate step from number 1.
 |      3)  Read the status register (port 388h).  Store the result.
 |      4)  Write FFh to register 2 (Timer 1).
 |      5)  Start timer 1 by writing 21h to register 4.
 |      6)  Delay for at least 80 microseconds.
 |      7)  Read the status register (port 388h).  Store the result.
 |      8)  Reset both timers and interrupts (see steps 1 and 2).
 |      9)  Test the stored results of steps 3 and 7 by ANDing them
 |          with E0h.  The result of step 3 should be 00h, and the 
 |          result of step 7 should be C0h.  If both are correct, an
 |          AdLib-compatible board is installed in the computer.
 */

   WriteOPL3Reg(0x04, 0x60);
   WriteOPL3Reg(0x04, 0x80);
   status1 = ReadOPL3Status();
   WriteOPL3Reg(0x02, 0xff);
   WriteOPL3Reg(0x04, 0x21);
   delay(1);
   status2 = ReadOPL3Status();
   WriteOPL3Reg(0x04, 0x60);
   WriteOPL3Reg(0x04, 0x80);
   if (!((status1 & 0xe0) == 0x00 && (status2 & 0xe0) == 0xc0))
      return OPN_NO_OPL3;

   /* Now detect OPL3 */
   status1 = ReadOPL3Status() & 0x06;
   if (status1 == 0x00)
      return OPN_SUCCESS;
   else
      return OPN_NO_OPL3;

#else

   return OPN_SUCCESS;

#endif
}

/*
 * Initialise everything.
 */
int OPN_Init(int type, int base, int num)
{

   /* check the arguments */
   switch(type)
   {
   case OPN_YM2203:
      if (num > 2) return OPN_TOOMANY;
      break;
   case OPN_YM2612: case OPN_YM2610:
      if (num > 1) return OPN_TOOMANY;
      break;
   }


   /* detect the SoundBlaster */
   OPL3_base = base;
   if (DetectOPL3() != OPN_SUCCESS) return OPN_NO_OPL3;
   OPN_initialised = 1;


   /* wake the SoundBlaster up */
   ResetOPL3();


   /* set everything up */
   num_OPN = num;
   OPN_type = type;
   YM[0].IRQHandler = NULL;
   YM[1].IRQHandler = NULL;
   YM[0].ch_offset = 0; YM[0].op_offset = 0;
   YM[1].ch_offset = 9; YM[1].op_offset = 18;

   OPN_Reset(0);
   if (OPN_type == OPN_YM2203) OPN_Reset(1);


   return OPN_SUCCESS;
}

/*
 * Set a handler to recieve debug messages.
 */
void OPN_SetDebugHandler(void (*func)(char *))
{
   msg_handler = func;
}

/*
 * Set the input clock for an OPN chip to the given value.
 */
void OPN_SetClock(int num, int hz)
{
   int d1,d2;

   YM[num].hertz = hz;
#ifdef SCALE_FREQUENCY
   /* see if OPN frequency is nearer to OPL or OPL/2 */

   d1 = abs(hz - OPL3_CLOCK / 2);
   d2 = abs(hz - OPL3_CLOCK);
   if (d1 < d2)
      YM[num].scale_factor = 1;
   else
      YM[num].scale_factor = 0;

   /* see if OPN frequency is nearer to OPL or OPL*2 */
   d1 = abs(hz - OPL3_CLOCK * 2);
   d2 = abs(hz - OPL3_CLOCK);
   if (d1 < d2) YM[num].scale_factor = -1;
#else
   YM[num].scale_factor = 0;
#endif

   /* chain through to the 2nd set */
   if (num == 0 && OPN_type != OPN_YM2203) OPN_SetClock(1, hz);
}

/*
 * Reset an OPN chip to it's power-on state.
 */
void OPN_Reset(int num)
{
   int n;

   /* clear all registers to zero */
   for (n=0;n<0x100;n++) OPN_WriteReg(&YM[num], n, 0x00);

   /* clear the 2nd set as well */
   if (OPN_type != OPN_YM2203)
      for (n=0;n<0x100;n++) OPN_WriteReg(&YM[1], n, 0x00);

   /* set the timers to be expired at start-up */
   YM[num].reg[0x00] = 0x80;

   YM[num].timer_a_event = YM[num].timer_b_event = NULL;
}

/*
 * Shut down the entire OPN emulator.
 * Resets the OPL3 as well.
 */
void OPN_Shutdown(void)
{
   int r;

   /* zero all registers: 1 to F5 (inclusive) */
   for (r=1;r<=0xf5;r++)
   {
      WriteOPL3Reg(r, 0x00);
   }

   /* set OPL3 mode */
   WriteOPL3Reg(0x105, 0x01);

   /* zero the 2nd register set as well */
   for (r=0x101;r<=0x1f5;r++)
   {
      /* leave OPL3 mode on for this bit */
      if (r == 0x105) continue;

      WriteOPL3Reg(r, 0x00);
   }

   /* un-set OPL3 mode */
   WriteOPL3Reg(0x105, 0x00);

   OPN_initialised = FALSE;
}

/*
 * This table maps an OPL3 operator to it's register offset.
 * Source: 
 *    "Programmer's Guide to Yamaha YMF 262/OPL3 FM Music Synthesizer,
 *     by Vladimir Arnost, QA-Software"
 */
static int op_addr[36] = { 0x000, 0x001, 0x002, 0x003, 0x004, 0x005,
                           0x008, 0x009, 0x00a, 0x00b, 0x00c, 0x00d,
                           0x010, 0x011, 0x012, 0x013, 0x014, 0x015,
                           0x100, 0x101, 0x102, 0x103, 0x104, 0x105,
                           0x108, 0x109, 0x10a, 0x10b, 0x10c, 0x10d,
                           0x110, 0x111, 0x112, 0x113, 0x114, 0x115 };

/*
 * This maps an OPL3 operator and base register, to the final OPL3 register.
 */
static DEF_INLINE int CalcOPAddr(int reg, int op)
{          
   return op_addr[op] + reg;
}

/*
 * This maps a OPN channel number (0-2) and operator (0-3), to a
 * OPL3 operator number.
 */
static DEF_INLINE int Calc_OPL3_OP(OPN *ym, int ch, int op)
{
   op *= 3;
   op += ch + ym->op_offset;
   return op;
}

/*
 * This maps a OPN channel (0-2) to it's OPL register base.
 */
static DEF_INLINE int CalcChannelAddr(OPN *ym, int ch)
{
   ch += ym->ch_offset;
   if (ch >= 9) ch += 0x100-9;
   return ch;
}

/*** NOTE:
 *
 * For all the following functions, arguments are assumed to be in the
 * OPN format, not OPL.
 *
 ***/

/*
 * Set which speakers a channel will come out of.
 */
static void SetStereo(OPN *ym, int ch, int stereo)
{
   int n;
   byte value;

   /* write to operators 1 & 2 */
   n = CalcChannelAddr(ym, ch) + 0xc0;
   value = ReadOPL3Reg(n);
   value &= ~0x30;
   if (stereo & 2) value |= 0x10;
   if (stereo & 1) value |= 0x20;
   WriteOPL3Reg(n, value);

   /* write to operators 3 & 4 */
   n = CalcChannelAddr(ym, ch+3) + 0xc0;
   value = ReadOPL3Reg(n);
   value &= ~0x30;
   if (stereo & 2) value |= 0x10;
   if (stereo & 1) value |= 0x20;
   WriteOPL3Reg(n, value);
}

/*
 * Do a KEY-ON for all the operators on the given OPN channel.
 * Problem: The OPN has separate control over them, the OPL doesn't.
 *          As long as the software controls them all at once, we're OK.
 */
static void KeyOn(OPN *ym, int ch)
{
   byte reg;
   int n;

   /* set to come out of left and right speakers, if mono */
   if (OPN_type == OPN_YM2203) SetStereo(ym, ch, 0x03);

   n = CalcChannelAddr(ym, ch) + 0xb0;
   reg = ReadOPL3Reg(n);
   reg |= 0x20;
   WriteOPL3Reg(n, reg);
}

/*
 * Do a KEY-OFF for all the operators on the given OPN channel.
 */
static void KeyOff(OPN *ym, int ch)
{
   byte reg;
   int n;
   
   n = CalcChannelAddr(ym, ch) + 0xb0;
   reg = ReadOPL3Reg(n);
   reg &= ~0x20;
   WriteOPL3Reg(n, reg);
}

/*
 * Set the F-Number and Block for the channel.
 * Block should have already been combined with F-Number, as one big number.
 */
static void SetFNumber(OPN *ym, int ch, int fn)
{
   int n;
   byte reg;

   n = CalcChannelAddr(ym, ch);
   reg = ReadOPL3Reg(0xb0+n);
   reg &= ~0x1f;

   fn >>= 1;
   WriteOPL3Reg(0xa0+n, (fn & 0xff));
   WriteOPL3Reg(0xb0+n, (reg | (fn >> 8)));
}

/*
 * This table maps the OPN connection algorithms to the OPL approximations.
 * (obtained by experiment)
 */
static int algorithm_table[8] = { 2, 1, 1, 1, 2, 3, 3, 3 };

/* Connection method #0 ideas so far:
 *
 *    Hmm.. it's tricky one. Method #0 seems to be prone to noise. (?)
 *    OPL #2 gives a clean sound, but doesn't really get all the tones right.
 *    OPL #1 works better, but too much static...
 *    OPL #0 *should* work perfectly... but doesn't? (noise)
 *    OPL #3 has no static, gets some tones right, but some very wrong...
 */

/*
 * Set the OPN connection algorithm for the channel.
 */
static void SetAlgorithm(OPN *ym, int ch, int alg)
{
   int n, syn1_2, syn3_4;
   byte value;

   alg = algorithm_table[alg];
   syn1_2 =  alg & 1;
   syn3_4 = (alg & 2) >> 1;

   /* Write the low bit to operators 1 & 2 */
   n = CalcChannelAddr(ym, ch) + 0xc0;
   value = ReadOPL3Reg(n);
   value &= ~0x01;
   WriteOPL3Reg(n, value | syn1_2);

   /* Write the high bit to operators 3 & 4 */
   n = CalcChannelAddr(ym, ch+3) + 0xc0;
   value = ReadOPL3Reg(n);
   value &= ~0x01;
   WriteOPL3Reg(n, value | syn3_4);
}

/*
 * Set the Frequency Multiplication Factor for an operator.
 */
static void SetMultiple(OPN *ym, int ch, int op, int mul)
{
   int reg;
   byte value;

#ifdef SCALE_FREQUENCY
   /* scale the value to account for the chip's frequency difference. */
   if (ym->scale_factor >= 0)
   {
      mul >>= ym->scale_factor;
   } else {
      mul <<= -(ym->scale_factor);
      if (mul == 0) mul = 1;
   }

   /* clip due to scaling */
   if (mul > 0x0f) mul = 0x0f;
   if (mul < 0x00) mul = 0x00;
#endif

   op = Calc_OPL3_OP(ym, ch, op);
   reg = CalcOPAddr(0x20, op);
   value = ReadOPL3Reg(reg);
   value &= 0xf0;
   WriteOPL3Reg(reg, value | mul);
}

/*
 * Set the OP-1 feedback for a channel.
 */
static void SetFeedback(OPN *ym, int ch, int fb)
{
   int reg;
   byte value;

   /* feedback - seems to need inverting for OPL. */
   if (fb != 0) fb ^= 0x07;

   fb <<= 1;
   reg = CalcChannelAddr(ym, ch) + 0xc0;
   value = ReadOPL3Reg(reg);
   value &= ~0x0e;
   WriteOPL3Reg(reg, value | fb);
}

/*
 * Set the total level (volume) for a channel.
 */
static void SetTotalLevel(OPN *ym, int ch, int op, int tl)
{
   int reg;
   byte value;

   op = Calc_OPL3_OP(ym, ch, op);
   reg = CalcOPAddr(0x40, op);
   value = ReadOPL3Reg(reg);
   value &= 0xc0;
   value |= tl >> 1;
   WriteOPL3Reg(reg, value);
}

/*
 * Set the Key Scale for a channel.
 */
static void SetKeyScale(OPN *ym, int ch, int op, int ks)
{
   int reg;
   byte value;

   op = Calc_OPL3_OP(ym, ch, op);
   ks <<= 3;
   ks &= 0x10;
   reg = CalcOPAddr(0x20, op);
   value = ReadOPL3Reg(reg);
   value &= ~0x10;
   WriteOPL3Reg(reg, value | ks);
}

/*
 * Set the attack rate for an operator.
 */
static void SetAttackRate(OPN *ym, int ch, int op, int ar)
{
   int reg;
   byte value;

   op = Calc_OPL3_OP(ym, ch, op);

#ifdef SCALE_FREQUENCY
   /* scale the value to account for the chip's frequency difference. */
   if (ym->scale_factor >= 0)
   {
      ar >>= ym->scale_factor;
   } else {
      ar <<= -(ym->scale_factor);
   }

   /* clip due to scaling */
   if (ar > 0x1f) ar = 0x1f;
   if (ar < 0x00) ar = 0x00;
#endif

   ar <<= 3;
   ar &= 0xf0;
   reg = CalcOPAddr(0x60, op);
   value = ReadOPL3Reg(reg);
   value &= 0x0f;
   WriteOPL3Reg(reg, value | ar);
}

/*
 * Set the decay rate for an operator.
 */
static void SetDecayRate(OPN *ym, int ch, int op, int dr)
{
   int reg;
   byte value;

   op = Calc_OPL3_OP(ym, ch, op);

#ifdef SCALE_FREQUENCY
   /* scale the value to account for the chip's frequency difference. */
   if (ym->scale_factor >= 0)
   {
      dr >>= ym->scale_factor;
   } else {
      dr <<= -(ym->scale_factor);
   }

   /* clip due to scaling */
   if (dr > 0x1f) dr = 0x1f;
   if (dr < 0x00) dr = 0x00;
#endif

   dr >>= 1;
   reg = CalcOPAddr(0x60, op);
   value = ReadOPL3Reg(reg);
   value &= 0xf0;
   WriteOPL3Reg(reg, value | dr);
}

/*
 * Set the sustain rate for an operator.
 */
static void SetSustainRate(OPN *ym, int ch, int op, int sr)
{
   int reg;
   byte value;

   op = Calc_OPL3_OP(ym, ch, op);
   reg = CalcOPAddr(0x20, op);
   value = ReadOPL3Reg(reg);
   value &= 0x20;

   /*
    * Here we have a problem. The OPL3 doesn't *have* a sustain rate.
    * It either sustains or doesn't.
    * So, if the value is lower than a threshold value, we sustain, otherwise
    * we don't.
    */

   if (sr < 0x18) value |= 0x20; /* 0x18 = 3/4ths of max. decay rate */
   WriteOPL3Reg(reg, value);
}

/*
 * Set the sustain level and release rate for an operator.
 * They should be combined together like so:
 *     sr_rr = (sr << 4) | rr;
 * (same format as the OPN register itself)
 */
static void SetSustainRelease(OPN *ym, int ch, int op, int sr_rr)
{
   int reg,rr;

#ifdef SCALE_FREQUENCY
   rr = sr_rr & 0x0f;

   /* scale the value to account for the chip's frequency difference. */
   if (ym->scale_factor >= 0)
   {
      rr >>= ym->scale_factor;
   } else {
      rr <<= -(ym->scale_factor);
   }

   /* clip due to scaling */
   if (rr > 0x0f) rr = 0x0f;
   if (rr < 0x00) rr = 0x00;
   sr_rr &= 0xf0;
   sr_rr |= rr;
#endif

   op = Calc_OPL3_OP(ym, ch, op);
   reg = CalcOPAddr(0x80, op);
   WriteOPL3Reg(reg, sr_rr);
}

/*
 * Write a value to a generic OPN register (FM part only).
 */
static void OPN_WriteReg(OPN *ym, byte reg, byte value)
{
   int op, ch, data;

   if (reg == 0x27 && (value & 0x40)) DebugMessage("Slot3!");

   switch(reg)
   {
   case 0x28: /* Key On/Off (SlotCh) */
      ch = value & 0x03;
      if (ch == 3) break;   /* only 3 channels, ignore 4th */
      if (value & 0xf0) KeyOn(ym, ch); else KeyOff(ym, ch);
      if ((value & 0xf0) != 0x00 && (value & 0xf0) != 0xf0)
         DebugMessage("Individual KEY-ON!");
      break;

   case 0x30:   case 0x31:   case 0x32:   case 0x33:
   case 0x34:   case 0x35:   case 0x36:   case 0x37:
   case 0x38:   case 0x39:   case 0x3a:   case 0x3b:
   case 0x3c:   case 0x3d:   case 0x3e:
							/* Detune / Multiple */
      ch = reg & 0x03;
      if (ch == 3) break;   /* only 3 channels, ignore 4th */
      op = (reg >> 2) & 3;
      SetMultiple(ym, ch, op, value & 0x0f);
      break;

   case 0x40:   case 0x41:   case 0x42:   case 0x43:
   case 0x44:   case 0x45:   case 0x46:   case 0x47:
   case 0x48:   case 0x49:   case 0x4a:   case 0x4b:
   case 0x4c:   case 0x4d:   case 0x4e:
							 /* Total Level (volume) */
      ch = reg & 0x03;
      if (ch == 3) break;   /* only 3 channels, ignore 4th */
      op = (reg >> 2) & 3;
      SetTotalLevel(ym, ch, op, value & 0x7f);
      break;

   case 0x50:   case 0x51:   case 0x52:   case 0x53:
   case 0x54:   case 0x55:   case 0x56:   case 0x57:
   case 0x58:   case 0x59:   case 0x5a:   case 0x5b:
   case 0x5c:   case 0x5d:   case 0x5e:
						 /* Key Scale / Attack Rate */
      ch = reg & 0x03;
      if (ch == 3) break;   /* only 3 channels, ignore 4th */
      op = (reg >> 2) & 3;
      SetAttackRate(ym, ch, op, value & 0x1f);
      SetKeyScale(ym, ch, op, value >> 6);
      break;

   case 0x60:   case 0x61:   case 0x62:   case 0x63:
   case 0x64:   case 0x65:   case 0x66:   case 0x67:
   case 0x68:   case 0x69:   case 0x6a:   case 0x6b:
   case 0x6c:   case 0x6d:   case 0x6e:
						 /* Decay Rate */
      ch = reg & 0x03;
      if (ch == 3) break;   /* only 3 channels, ignore 4th */
      op = (reg >> 2) & 3;
      SetDecayRate(ym, ch, op, value & 0x1f);
      break;

   case 0x70:   case 0x71:   case 0x72:   case 0x73:
   case 0x74:   case 0x75:   case 0x76:   case 0x77:
   case 0x78:   case 0x79:   case 0x7a:   case 0x7b:
   case 0x7c:   case 0x7d:   case 0x7e:
					/* Sustain Rate */
      ch = reg & 0x03;
      if (ch == 3) break;   /* only 3 channels, ignore 4th */
      op = (reg >> 2) & 3;
      SetSustainRate(ym, ch, op, value & 0x1f);
      break;

   case 0x80:   case 0x81:   case 0x82:   case 0x83:
   case 0x84:   case 0x85:   case 0x86:   case 0x87:
   case 0x88:   case 0x89:   case 0x8a:   case 0x8b:
   case 0x8c:   case 0x8d:   case 0x8e:
					 /* Sustain Level / Release Rate */
      ch = reg & 0x03;
      if (ch == 3) break;   /* only 3 channels, ignore 4th */
      op = (reg >> 2) & 3;
      value |= 0x0f;
      SetSustainRelease(ym, ch, op, value);
      break;

   case 0xa0:   case 0xa1:   case 0xa2:
						 /* F-Number */
      data = value | ((ym->reg[reg+4] & 0x3f) << 8);
      SetFNumber(ym, reg - 0xa0, data);
      break;

   case 0xa4:   case 0xa5:   case 0xa6:
							/* F-Number / Block */
      data = ym->reg[reg-4] | ((value & 0x3f) << 8);
      SetFNumber(ym, reg - 0xa4, data);
      break;

   case 0xb0:   case 0xb1:   case 0xb2:
					 /* Feedback / Algorithm */
      /* Not all connection-algorithms can be done on OPL,
       * so just try and merge similiar ones.
       */
      SetAlgorithm(ym, reg - 0xb0, value & 7);
      SetFeedback(ym, reg - 0xb0, (value >> 3) & 7);
      break;

   case 0xb4:   case 0xb5:   case 0xb6:
					 /* L+R / AMS / FMS */
      if (OPN_type != OPN_YM2203)
         SetStereo(ym, reg - 0xb4, value >> 6);
      break;
   }
}

/*
 * Read the status register for a given chip.
 */
byte OPN_ReadStatus(int num)
{
   return YM[num].reg[0x00];
}

/*
 * Write a value to a YM2203 register.
 */
void YM2203_WriteReg(int num, int addr, byte value)
{
   OPN *ym;
   byte reg;
   ym = &YM[num];

   if (addr == 0)
   {
      ym->reg_latch = value;
   } else {
#ifdef ONLY_PLAY_CHANNEL
      if (reg == 0x28) /* Remove KEY-ON's except that channel */
      {
         int ch;
         ch = ONLY_PLAY_CHANNEL - 1;
         if ((value & 3) != ch) return;
      }
#endif

      reg = ym->reg_latch;
      ym->reg[reg] = value;
      OPN_WriteReg(ym, reg, value);
   }
}

/*
 * Write a value to a YM2610 register.
 */
void YM2610_WriteReg(int addr, byte value)
{
   OPN *ym;
   byte reg;

   if (addr & 0x02)  /* 2nd section */
      ym = &YM[1];
   else              /* 1st section */
      ym = &YM[0];

   if (addr & 0x01)
   {
      reg = ym->reg_latch;
      ym->reg[reg] = value;

#ifdef ONLY_PLAY_CHANNEL
      if (reg == 0x28) /* Remove KEY-ON's except that channel */
      {
         int ch;
         ch = ONLY_PLAY_CHANNEL;
         if (ch < 4) ch--;
         if ((value & 7) != ch) return;
      }
#endif

      if (addr == 1) switch(reg)
      {
      case 0x28: /* Key On/Off (SlotCh) */
         /* If doing KEY-ON to channels 4-6, switch to 2nd register set */
         if (value & 0x04)
         {
            ym = &YM[1];
            value &= ~0x04;
         }
         break;
      }

      OPN_WriteReg(ym, reg, value);
   } else {
      ym->reg_latch = value;
   }
}

/*
 * Write a value to a YM2612 register.
 */
void YM2612_WriteReg(int addr, byte value)
{
   OPN *ym;
   byte reg;

   if (addr & 0x02)  /* 2nd section */
      ym = &YM[1];
   else              /* 1st section */
      ym = &YM[0];

   if (addr & 0x01)
   {
      reg = ym->reg_latch;
      ym->reg[reg] = value;

#ifdef ONLY_PLAY_CHANNEL
      if (reg == 0x28) /* Remove KEY-ON's except that channel */
      {
         int ch;
         ch = ONLY_PLAY_CHANNEL;
         if (ch < 4) ch--;
         if ((value & 7) != ch) return;
      }
#endif

      if (addr == 1) switch(reg)
      {
      case 0x28: /* Key On/Off (SlotCh) */
         /* If doing KEY-ON to channels 4-6, switch to 2nd register set */
         if (value & 0x04)
         {
            ym = &YM[1];
            value &= ~0x04;
         }
         break;

      case 0x2a: /* DAC data */
         if (YM[0].reg[0x2b] & 0x80) /* check for DAC enable */
         {
            DebugMessage("DAC write");
         }
         break;

      case 0x2b: /* DAC enable */
         if (value & 0x80) DebugMessage("DAC enabled");
         break;
      }

      OPN_WriteReg(ym, reg, value);
   } else {
      ym->reg_latch = value;
   }
}
