/******************************************************************************/
/*                                                                            */
/*                          RAINE COLOUR MAPPING                              */
/*                                                                            */
/******************************************************************************/

#include <limits.h>
#include "raine.h"
#include "gameinc.h"
#include "palette.h"
#include "blitasm.h"

/*

Closest Colour finder (when there are no free pens)

*/

static void closest_colour_init(void);
static UBYTE closest_colour(int r, int g, int b);

/*

Data

*/

enum map_type
{
   PMAP_NONE = 0,
   PMAP_MAPPED,
   PMAP_DIRECT,
};

static UBYTE map_mode = PMAP_NONE;

static UBYTE *RAM_PAL;		// points to source colour ram

static UBYTE *colour_ram_src[MAX_COLBANKS];

static UBYTE *bankmap;
static UBYTE *cloc;

static ULONG res_pens;		// number of pens reserved in the palette (0-255)
static ULONG talc;		// next free pen in the palette (0-255)

static UBYTE *cloc_2;		// cloc data from last frame
static UBYTE *pmap;		// list of which pens are allocated (0=free 1=used)
static UBYTE *pmap_2;		// pmap data from last frame
static UWORD pal_banks;		// number of colour banks the hardware supports
static ULONG map_size;

#ifdef TRIPLE_BUFFER
static UBYTE *cloc_2a;		// extra buffer
static UBYTE *pmap_2a;		// extra buffer
#else
#define cloc_2a	cloc_2
#define pmap_2a	pmap_2
#endif

static PALETTE pal_screen;

static UBYTE white_pen;

void InitPaletteMap(UBYTE *src, int banks, int bankcols, int mapsize)
{
   int ta;

   map_mode = PMAP_MAPPED;

   RAM_PAL = src;

   pal_banks = banks;
   map_size  = mapsize;

   bankmap = AllocateMem(pal_banks * bankcols * ((display_cfg.bpp + 7) / 8));
   cloc    = AllocateMem(map_size);
   cloc_2  = AllocateMem(map_size);
   #ifdef TRIPLE_BUFFER
   cloc_2a = AllocateMem(map_size);
   #endif
   pmap    = AllocateMem(256);
   pmap_2  = AllocateMem(256);
   #ifdef TRIPLE_BUFFER
   pmap_2a = AllocateMem(256);
   #endif

   for(ta=0; ta<pal_banks; ta++){
      coltab[ta] = bankmap + (ta * bankcols * ((display_cfg.bpp + 7) / 8));
      colour_ram_src[ta] = src + (ta * bankcols * 2);
   }

   res_pens = 1;	// 1 reserved (pen 0 == black)

   white_pen = 255;

   closest_colour_init();

   reset_palette_map();
}

void set_colour_mapper(COLOUR_MAPPER *colour_mapper)
{
   switch(display_cfg.bpp){
      case 8:
         current_mapper = colour_mapper->mapper_8bpp;
      break;
      case 15:
         current_mapper = colour_mapper->mapper_15bpp;
      break;
      case 16:
         current_mapper = colour_mapper->mapper_16bpp;
      break;
      case 24:
         current_mapper = colour_mapper->mapper_24bpp;
      break;
      case 32:
         current_mapper = colour_mapper->mapper_32bpp;
      break;
      default:
         current_mapper = colour_mapper->mapper_8bpp;
      break;
   }
}

void set_white_pen(int pen)
{
   white_pen = pen;
}

UBYTE get_white_pen(void)
{
   pal[white_pen].r = 63;
   pal[white_pen].g = 63;
   pal[white_pen].b = 63;

   return white_pen;
}

int Reserve_Pens(int num)
{
   int ret;

   ret = res_pens;
   res_pens += num;

   return ret;
}

void Set_Pens_15bit_xRGB(int talc, int start, int cols)
{
   UWORD yy;
   UWORD *ta;

   ta = (UWORD *) (RAM_PAL+(start<<1));
   do{
      yy = (*ta++) & 0x7FFF;
      pal[talc].r = (yy&0x7C00)>>9;
      pal[talc].g = (yy&0x03E0)>>4;
      pal[talc].b = (yy&0x001F)<<1;
      talc++;
   }while(--cols);
}

void Set_Pens_12bit_RGBx(int talc, int start, int cols)
{
   UWORD yy;
   UWORD *ta;

   ta = (UWORD *) (RAM_PAL+(start<<1));
   do{
      yy = (*ta++) >> 4;
      pal[talc].r = (yy&0x0F00)>>6;
      pal[talc].g = (yy&0x00F0)>>2;
      pal[talc].b = (yy&0x000F)<<2;
      talc++;
   }while(--cols);
}

void ClearPaletteMap(void)
{
   UBYTE *ta;

   switch(map_mode){
   case PMAP_MAPPED:

   #ifdef TRIPLE_BUFFER

   ta      = cloc_2;
   cloc_2  = cloc_2a;		// cloc_2  = <last-last frame map>
   cloc_2a = cloc;		// cloc_2a = <last frame map>
   cloc    = ta;		// cloc    = <new frame map>

   ta      = pmap_2;
   pmap_2  = pmap_2a;		// pmap_2  = <last-last frame map>
   pmap_2a = pmap;		// pmap_2a = <last frame map>
   pmap    = ta;		// pmap    = <new frame map>

   #else
   ta     = cloc_2;
   cloc_2 = cloc;		// cloc_2 = <last frame map>
   cloc   = ta;			// cloc   = <new frame map>

   ta     = pmap_2;
   pmap_2 = pmap;		// pmap_2 = <last frame map>
   pmap   = ta;			// pmap   = <new frame map>
   #endif

   memset(bank_status, 0, pal_banks * 2);
   memset(cloc, 0xFF, map_size);
   memset(pmap, 0x00, 256);

   // pen 0 is always black

   cloc[0] = 0;
   pal[0].r = 0;
   pal[0].g = 0;
   pal[0].b = 0;

   // Find first free pen

   talc = res_pens;

   #ifdef TRIPLE_BUFFER
   while((pmap_2[talc])||(pmap_2a[talc])){ talc++; }
   #else
   talc --;
   while(pmap_2[++talc]);
   #endif

   break;
   case PMAP_DIRECT:
   break;
   default:

   break;
   }
}

void reset_palette_map(void)
{
   switch(map_mode){
   case PMAP_MAPPED:
      ClearPaletteMap();		// Make sure no colours are cached
      ClearPaletteMap();
      #ifdef TRIPLE_BUFFER
      ClearPaletteMap();
      #endif
   break;
   case PMAP_DIRECT:
   break;
   default:

   break;
   }

   memset(pal       , 0x00, sizeof(PALETTE));
   memset(pal_screen, 0xFF, sizeof(PALETTE));
}

void destroy_palette_map(void)
{
   map_mode = PMAP_NONE;
}

void silly_wgp_colour_ram(UBYTE *src)
{
   RAM_PAL = src;
}

#define GET_PEN_FOR_COLOUR_8(red, green, blue, result)                       \
      if(cloc_2a[yy] == 255){                                                \
         if(talc < 255){                                                     \
            result = cloc_2a[yy] = cloc[yy] = talc;                          \
            pal[talc].r = red;                                               \
            pal[talc].g = green;                                             \
            pal[talc].b = blue;                                              \
            talc++;                                                          \
            while((pmap_2[talc])||(pmap_2a[talc])){ talc++; }                \
         }                                                                   \
         else{				                                     \
            result = closest_colour(red, green, blue);                       \
         }                                                                   \
      }                                                                      \
      else{                                                                  \
         result = cloc[yy] = cloc_2a[yy];                                    \
      }                                                                      \
      pmap[result] = 1;			                                     \

#define GET_PEN_FOR_COLOUR_15(red, green, blue, result)                      \
      result = (((red) << 9) | ((green) << 4) | ((blue) >> 1)) & 0x7FFF;     \

#define GET_PEN_FOR_COLOUR_16(red, green, blue, result)                      \
      result = (((red) << 10) | ((green) << 5) | ((blue) >> 1)) & 0xFFFF;    \

#define GET_PEN_FOR_COLOUR_24(red, green, blue, result)                      \
      result = (((red) << 18) | ((green) << 10) | ((blue) << 2)) & 0xFFFFFF; \

#define GET_PEN_FOR_COLOUR_32(red, green, blue, result)                      \
      result = (((red) << 18) | ((green) << 10) | ((blue) << 2)) & 0xFFFFFF; \

/******************************************************************************/
/* 12-bit xBGR (12-bit mapping)                                               */
/******************************************************************************/

void Map_12bit_xRGB(int bank, int cols)
{
   UWORD yy;
   UWORD *ta;
   UBYTE *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5));
   ct = coltab[bank];
   do{
      yy = (*ta++) & 0x0FFF;

      GET_PEN_FOR_COLOUR_8(
         (yy&0x0F00)>>6,
         (yy&0x00F0)>>2,
         (yy&0x000F)<<2,
         res
      );

      *ct++ = res;

   }while(--cols);
}

void map_12bit_xxxx_rrrr_gggg_bbbb_8(ULONG bank, ULONG cols)
{
   UWORD yy;
   UWORD *ta;
   UBYTE *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5));
   ct = coltab[bank];
   do{
      yy = (*ta++) & 0x0FFF;

      GET_PEN_FOR_COLOUR_8(
         (yy&0x0F00)>>6,
         (yy&0x00F0)>>2,
         (yy&0x000F)<<2,
         res
      );

      *ct++ = res;

   }while(--cols);
}

void map_12bit_xxxx_rrrr_gggg_bbbb_15(ULONG bank, ULONG cols)
{
   UWORD yy;
   UWORD *ta;
   UWORD *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5));
   ct = (UWORD *) coltab[bank];
   do{
      yy = (*ta++) & 0x0FFF;

      GET_PEN_FOR_COLOUR_15(
         (yy&0x0F00)>>6,
         (yy&0x00F0)>>2,
         (yy&0x000F)<<2,
         res
      );

      *ct++ = res;

   }while(--cols);
}

struct COLOUR_MAPPER col_map_xxxx_rrrr_gggg_bbbb =
{
   "12bit xxxx rrrr gggg bbbb",
   map_12bit_xxxx_rrrr_gggg_bbbb_8,
   map_12bit_xxxx_rrrr_gggg_bbbb_15,
};

void Map_12bit_xxxxRRRRGGGGBBBB_Rev(int bank, int cols)
{
   UWORD yy;
   UWORD *ta;
   UBYTE *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5)+30);
   ct = coltab[bank];
   do{
      yy = (*ta--) & 0x0FFF;

      GET_PEN_FOR_COLOUR_8(
         (yy&0x0F00)>>6,
         (yy&0x00F0)>>2,
         (yy&0x000F)<<2,
         res
      );

      *ct++ = res;

   }while(--cols);
}

/******************************************************************************/
/* 12-bit xBGR (12-bit mapping)                                               */
/******************************************************************************/

void Map_12bit_xBGR(int bank, int cols)
{
   UWORD yy;
   UWORD *ta;
   UBYTE *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5));
   ct = coltab[bank];
   do{
      yy = (*ta++) & 0x0FFF;

      GET_PEN_FOR_COLOUR_8(
         (yy&0x000F)<<2,
         (yy&0x00F0)>>2,
         (yy&0x0F00)>>6,
         res
      );

      *ct++ = res;

   }while(--cols);
}

void Map_12bit_xBGR_wf(int bank, int cols)
{
   UWORD yy;
   UWORD *ta;
   UBYTE *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<7));		// <<7
   ct = coltab[bank];
   do{
      yy = (*ta++) & 0x0FFF;

      GET_PEN_FOR_COLOUR_8(
         (yy&0x000F)<<2,
         (yy&0x00F0)>>2,
         (yy&0x0F00)>>6,
         res
      );

      *ct++ = res;

   }while(--cols);
}

/******************************************************************************/
/* 12-bit RGBx (12-bit mapping)                                               */
/******************************************************************************/

void Map_12bit_RGBx(int bank, int cols)
{
   UWORD yy;
   UWORD *ta;
   UBYTE *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5));
   ct = coltab[bank];
   do{
      yy = (*ta++) >> 4;

      GET_PEN_FOR_COLOUR_8(
         (yy&0x0F00)>>6,
         (yy&0x00F0)>>2,
         (yy&0x000F)<<2,
         res
      );

      *ct++ = res;

   }while(--cols);
}

/******************************************************************************/
/* 15-bit RRRRGGGGBBBBRGBx Reverse Colour Order (Jaleco MS-1)                 */
/******************************************************************************/

void Map_15bit_RRRRGGGGBBBBRGBx_Rev(int bank, int cols)
{
   UWORD yy;
   UWORD *ta;
   UBYTE *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5)+30);
   ct = coltab[bank];
   do{
      yy = (*ta--) >> 1;

      GET_PEN_FOR_COLOUR_8(
         ((yy&0x7800)>>9) | ((yy&0x0004)>>1),
         ((yy&0x0780)>>5) | ((yy&0x0002)>>0),
         ((yy&0x0078)>>1) | ((yy&0x0001)<<1),
         res
      );

      *ct++ = res;

   }while(--cols);
}

void map_15bit_rrrr_gggg_bbbb_rgbx_rev_8(ULONG bank, ULONG cols)
{
   UWORD yy;
   UWORD *ta;
   UBYTE *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5)+30);
   ct = coltab[bank];
   do{
      yy = (*ta--) >> 1;

      GET_PEN_FOR_COLOUR_8(
         ((yy&0x7800)>>9) | ((yy&0x0004)>>1),
         ((yy&0x0780)>>5) | ((yy&0x0002)>>0),
         ((yy&0x0078)>>1) | ((yy&0x0001)<<1),
         res
      );

      *ct++ = res;

   }while(--cols);
}

void map_15bit_rrrr_gggg_bbbb_rgbx_rev_15(ULONG bank, ULONG cols)
{
   UWORD yy;
   UWORD *ta;
   UWORD *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5)+30);
   ct = (UWORD *) coltab[bank];
   do{
      yy = (*ta--) >> 1;

      GET_PEN_FOR_COLOUR_15(
         ((yy&0x7800)>>9) | ((yy&0x0004)>>1),
         ((yy&0x0780)>>5) | ((yy&0x0002)>>0),
         ((yy&0x0078)>>1) | ((yy&0x0001)<<1),
         res
      );

      *ct++ = res;

   }while(--cols);
}

struct COLOUR_MAPPER col_map_rrrr_gggg_bbbb_rgbx_rev =
{
   "15bit rrrr gggg bbbb rgbx [reverse]",
   map_15bit_rrrr_gggg_bbbb_rgbx_rev_8,
   map_15bit_rrrr_gggg_bbbb_rgbx_rev_15,
};

void Map_15bit_RRRRGGGGBBBBRGBx(int bank, int cols)
{
   UWORD yy;
   UWORD *ta;
   UBYTE *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5));
   ct = coltab[bank];
   do{
      yy = (*ta++) >> 1;

      GET_PEN_FOR_COLOUR_8(
         ((yy&0x7800)>>9) | ((yy&0x0004)>>1),
         ((yy&0x0780)>>5) | ((yy&0x0002)>>0),
         ((yy&0x0078)>>1) | ((yy&0x0001)<<1),
         res
      );

      *ct++ = res;

   }while(--cols);
}

void Map_15bit_RRRRRGGGGGBBBBBx_Rev(int bank, int cols)
{
   UWORD yy;
   UWORD *ta;
   UBYTE *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5)+30);
   ct = coltab[bank];
   do{
      yy = (*ta--) >> 1;

      GET_PEN_FOR_COLOUR_8(
         (yy&0x7C00)>>9,
         (yy&0x03E0)>>4,
         (yy&0x001F)<<1,
         res
      );

      *ct++ = res;

   }while(--cols);
}


void map_15bit_rrrr_rggg_ggbb_bbbx_rev_8(ULONG bank, ULONG cols)
{
   UWORD yy;
   UWORD *ta;
   UBYTE *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5)+30);
   ct = coltab[bank];
   do{
      yy = (*ta--) >> 1;

      GET_PEN_FOR_COLOUR_8(
         (yy&0x7C00)>>9,
         (yy&0x03E0)>>4,
         (yy&0x001F)<<1,
         res
      );

      *ct++ = res;

   }while(--cols);
}

void map_15bit_rrrr_rggg_ggbb_bbbx_rev_15(ULONG bank, ULONG cols)
{
   UWORD yy;
   UWORD *ta;
   UWORD *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5)+30);
   ct = (UWORD *) coltab[bank];
   do{
      yy = (*ta--) >> 1;

      GET_PEN_FOR_COLOUR_15(
         (yy&0x7C00)>>9,
         (yy&0x03E0)>>4,
         (yy&0x001F)<<1,
         res
      );

      *ct++ = res;

   }while(--cols);
}

struct COLOUR_MAPPER col_map_rrrr_rggg_ggbb_bbbx_rev =
{
   "15bit rrrr rggg ggbb bbbx [reverse]",
   map_15bit_rrrr_rggg_ggbb_bbbx_rev_8,
   map_15bit_rrrr_rggg_ggbb_bbbx_rev_15,
};

/******************************************************************************/
/* 12-bit xBGR (12-bit mapping) Reverse Colour Order (WestStory)              */
/******************************************************************************/

void Map_12bit_xBGR_Rev(int bank, int cols)
{
   UWORD yy;
   UWORD *ta;
   UBYTE *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5)+30);
   ct = coltab[bank];
   do{
      yy = (*ta--) & 0x0FFF;

      GET_PEN_FOR_COLOUR_8(
         (yy&0x000F)<<2,
         (yy&0x00F0)>>2,
         (yy&0x0F00)>>6,
         res
      );

      *ct++ = res;

   }while(--cols);
}

/******************************************************************************/
/* 12-bit RGBx (12-bit mapping)                                               */
/******************************************************************************/

void map_12bit_rrrr_gggg_bbbb_xxxx_8(ULONG bank, ULONG cols)
{
   UWORD yy;
   UWORD *ta;
   UBYTE *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5)+30);
   ct = coltab[bank];
   do{
      yy = ReadWord68k(ta--) >> 4;

      GET_PEN_FOR_COLOUR_8(
         (yy&0x0F00)>>6,
         (yy&0x00F0)>>2,
         (yy&0x000F)<<2,
         res
      );

      *ct++ = res;

   }while(--cols);
}

void map_12bit_rrrr_gggg_bbbb_xxxx_15(ULONG bank, ULONG cols)
{
   UWORD yy;
   UWORD *ta;
   UWORD *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5)+30);
   ct = (UWORD *) coltab[bank];
   do{
      yy = ReadWord68k(ta--) >> 4;

      GET_PEN_FOR_COLOUR_15(
         (yy&0x0F00)>>6,
         (yy&0x00F0)>>2,
         (yy&0x000F)<<2,
         res
      );

      *ct++ = res;

   }while(--cols);
}

struct COLOUR_MAPPER col_map_rrrr_gggg_bbbb_xxxx_68k =
{
   "12bit rrrr gggg bbbb xxxx [M68000]",
   map_12bit_rrrr_gggg_bbbb_xxxx_8,
   map_12bit_rrrr_gggg_bbbb_xxxx_15,
};

/******************************************************************************/
/* 15-bit xRGB (15-bit mapping)                                               */
/******************************************************************************/

void Map_15bit_xRGB(int bank, int cols)
{
   UWORD yy;
   UWORD *ta;
   UBYTE *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5));
   ct = coltab[bank];
   do{
      yy = (*ta++) & 0x7FFF;

      GET_PEN_FOR_COLOUR_8(
         (yy&0x7C00)>>9,
         (yy&0x03E0)>>4,
         (yy&0x001F)<<1,
         res
      );

      *ct++ = res;

   }while(--cols);
}

void map_15bit_xrrr_rrgg_gggb_bbbb_8(ULONG bank, ULONG cols)
{
   UWORD yy;
   UWORD *ta;
   UBYTE *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5));
   ct = coltab[bank];
   do{
      yy = (*ta++) & 0x7FFF;

      GET_PEN_FOR_COLOUR_8(
         (yy&0x7C00)>>9,
         (yy&0x03E0)>>4,
         (yy&0x001F)<<1,
         res
      );

      *ct++ = res;

   }while(--cols);
}

void map_15bit_xrrr_rrgg_gggb_bbbb_15(ULONG bank, ULONG cols)
{
   UWORD yy;
   UWORD *ta;
   UWORD *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5));
   ct = (UWORD *) coltab[bank];
   do{
      yy = (*ta++) & 0x7FFF;

      GET_PEN_FOR_COLOUR_15(
         (yy&0x7C00)>>9,
         (yy&0x03E0)>>4,
         (yy&0x001F)<<1,
         res
      );

      *ct++ = res;

   }while(--cols);
}

void map_15bit_xrrr_rrgg_gggb_bbbb_16(ULONG bank, ULONG cols)
{
   UWORD yy;
   UWORD *ta;
   UWORD *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5));
   ct = (UWORD *) coltab[bank];
   do{
      yy = (*ta++) & 0x7FFF;

      GET_PEN_FOR_COLOUR_16(
         (yy&0x7C00)>>9,
         (yy&0x03E0)>>4,
         (yy&0x001F)<<1,
         res
      );

      *ct++ = res;

   }while(--cols);
}

void map_15bit_xrrr_rrgg_gggb_bbbb_24(ULONG bank, ULONG cols)
{
   UWORD yy;
   UWORD *ta;
   ULONG *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5));
   ct = (UWORD *) coltab[bank];
   do{
      yy = (*ta++) & 0x7FFF;

      GET_PEN_FOR_COLOUR_24(
         (yy&0x7C00)>>9,
         (yy&0x03E0)>>4,
         (yy&0x001F)<<1,
         res
      );

      *ct++ = res;

   }while(--cols);
}

void map_15bit_xrrr_rrgg_gggb_bbbb_32(ULONG bank, ULONG cols)
{
   UWORD yy;
   UWORD *ta;
   ULONG *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5));
   ct = (UWORD *) coltab[bank];
   do{
      yy = (*ta++) & 0x7FFF;

      GET_PEN_FOR_COLOUR_32(
         (yy&0x7C00)>>9,
         (yy&0x03E0)>>4,
         (yy&0x001F)<<1,
         res
      );

      *ct++ = res;

   }while(--cols);
}

struct COLOUR_MAPPER col_map_xrrr_rrgg_gggb_bbbb =
{
   "15bit xrrr rrgg gggb bbbb",
   map_15bit_xrrr_rrgg_gggb_bbbb_8,
   map_15bit_xrrr_rrgg_gggb_bbbb_15,
   map_15bit_xrrr_rrgg_gggb_bbbb_16,
   map_15bit_xrrr_rrgg_gggb_bbbb_24,
   map_15bit_xrrr_rrgg_gggb_bbbb_32,
};

/******************************************************************************/
/* 15-bit xBGR (15-bit mapping)                                               */
/******************************************************************************/

void Map_15bit_xBGR(int bank, int cols)
{
   UWORD yy;
   UWORD *ta;
   UBYTE *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5));
   ct = coltab[bank];
   do{
      yy = (*ta++) & 0x7FFF;

      GET_PEN_FOR_COLOUR_8(
         (yy&0x001F)<<1,
         (yy&0x03E0)>>4,
         (yy&0x7C00)>>9,
         res
      );

      *ct++ = res;

   }while(--cols);
}

void map_15bit_xbbb_bbgg_gggr_rrrr_8(ULONG bank, ULONG cols)
{
   UWORD yy;
   UWORD *ta;
   UBYTE *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5));
   ct = coltab[bank];
   do{
      yy = (*ta++) & 0x7FFF;

      GET_PEN_FOR_COLOUR_8(
         (yy&0x001F)<<1,
         (yy&0x03E0)>>4,
         (yy&0x7C00)>>9,
         res
      );

      *ct++ = res;

   }while(--cols);
}

void map_15bit_xbbb_bbgg_gggr_rrrr_15(ULONG bank, ULONG cols)
{
   UWORD yy;
   UWORD *ta;
   UWORD *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5));
   ct = (UWORD *) coltab[bank];
   do{
      yy = (*ta++) & 0x7FFF;

      GET_PEN_FOR_COLOUR_15(
         (yy&0x001F)<<1,
         (yy&0x03E0)>>4,
         (yy&0x7C00)>>9,
         res
      );

      *ct++ = res;

   }while(--cols);
}

struct COLOUR_MAPPER col_map_xbbb_bbgg_gggr_rrrr =
{
   "15bit xbbb bbgg gggr rrrr",
   map_15bit_xbbb_bbgg_gggr_rrrr_8,
   map_15bit_xbbb_bbgg_gggr_rrrr_15,
};

/******************************************************************************/
/* 15-bit xBGR (15-bit mapping) [only 4bits used] AIR-SYSTEM                  */
/******************************************************************************/

void Map_15bit_xBGR_4bit(int bank, int cols)
{
   UWORD yy;
   UWORD *ta;
   UBYTE *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) (RAM_PAL+(bank<<5));
   ct = coltab[bank];
   do{
      yy = (*ta++) & 0x7FFF;

      GET_PEN_FOR_COLOUR_8(
         (yy&0x000F)<<2,
         (yy&0x01E0)>>3,
         (yy&0x3C00)>>8,
         res
      );

      *ct++ = res;

   }while(--cols);
}

/******************************************************************************/
/* 24-bit xRGB (12-bit mapping) 680x0 format                                  */
/******************************************************************************/

void Map_24bit_xRGB(int bank, int cols)
{
   UBYTE yr,yg,yb;
   UWORD yy;
   UBYTE *ta;
   UBYTE *ct,res;

   bank_status[bank] = cols;
   ta = (RAM_PAL+(bank<<6));
   ct = coltab[bank];
   do{
      yr = (*(ta+1)) >>4;
      yg = (*(ta+2)) >>4;
      yb = (*(ta+3)) >>4;
      ta += 4;
      yy=((yr<<8)|(yg<<4)|(yb));

      GET_PEN_FOR_COLOUR_8(
         yr<<2,
         yg<<2,
         yb<<2,
         res
      );

      *ct++ = res;

   }while(--cols);
}

void map_24bit_xxxx_xxxx_rrrr_rrrr_gggg_gggg_bbbb_bbbb_8(ULONG bank, ULONG cols)
{
   UBYTE yr,yg,yb;
   ULONG yy;
   ULONG *ta;
   UBYTE *ct,res;

   bank_status[bank] = cols;
   ta = (ULONG *) (RAM_PAL+(bank<<6));
   ct = coltab[bank];
   do{
      yy = (*ta++) >> 8;

      yr = (yy&0x0000F8)>>3;
      yg = (yy&0x00F800)>>11;
      yb = (yy&0xF80000)>>19;

      yy = (yr<<10) | (yg<<5) | (yb<<0);

      GET_PEN_FOR_COLOUR_8(
         yr << 1,
         yg << 1,
         yb << 1,
         res
      );

      *ct++ = res;

   }while(--cols);
}

void map_24bit_xxxx_xxxx_rrrr_rrrr_gggg_gggg_bbbb_bbbb_15(ULONG bank, ULONG cols)
{
   UBYTE yr,yg,yb;
   ULONG yy;
   ULONG *ta;
   UWORD *ct,res;

   bank_status[bank] = cols;
   ta = (ULONG *) (RAM_PAL+(bank<<6));
   ct = (UWORD *) coltab[bank];
   do{
      yy = (*ta++) >> 8;

      yr = (yy&0x0000F8)>>3;
      yg = (yy&0x00F800)>>11;
      yb = (yy&0xF80000)>>19;

      yy = (yr<<10) | (yg<<5) | (yb<<0);

      GET_PEN_FOR_COLOUR_15(
         yr << 1,
         yg << 1,
         yb << 1,
         res
      );

      *ct++ = res;

   }while(--cols);
}

void map_24bit_xxxx_xxxx_rrrr_rrrr_gggg_gggg_bbbb_bbbb_16(ULONG bank, ULONG cols)
{
   UBYTE yr,yg,yb;
   ULONG yy;
   ULONG *ta;
   UWORD *ct,res;

   bank_status[bank] = cols;
   ta = (ULONG *) (RAM_PAL+(bank<<6));
   ct = (UWORD *) coltab[bank];
   do{
      yy = (*ta++) >> 8;

      yr = (yy&0x0000F8)>>3;
      yg = (yy&0x00F800)>>11;
      yb = (yy&0xF80000)>>19;

      yy = (yr<<10) | (yg<<5) | (yb<<0);

      GET_PEN_FOR_COLOUR_16(
         yr << 1,
         yg << 1,
         yb << 1,
         res
      );

      *ct++ = res;

   }while(--cols);
}

void map_24bit_xxxx_xxxx_rrrr_rrrr_gggg_gggg_bbbb_bbbb_24(ULONG bank, ULONG cols)
{
   UBYTE yr,yg,yb;
   ULONG yy;
   ULONG *ta;
   ULONG *ct,res;

   bank_status[bank] = cols;
   ta = (ULONG *) (RAM_PAL+(bank<<6));
   ct = (UWORD *) coltab[bank];
   do{
      yy = (*ta++) >> 8;

      yr = (yy&0x0000F8)>>3;
      yg = (yy&0x00F800)>>11;
      yb = (yy&0xF80000)>>19;

      yy = (yr<<10) | (yg<<5) | (yb<<0);

      GET_PEN_FOR_COLOUR_24(
         yr << 1,
         yg << 1,
         yb << 1,
         res
      );

      *ct++ = res;

   }while(--cols);
}

void map_24bit_xxxx_xxxx_rrrr_rrrr_gggg_gggg_bbbb_bbbb_32(ULONG bank, ULONG cols)
{
   UBYTE yr,yg,yb;
   ULONG yy;
   ULONG *ta;
   ULONG *ct,res;

   bank_status[bank] = cols;
   ta = (ULONG *) (RAM_PAL+(bank<<6));
   ct = (UWORD *) coltab[bank];
   do{
      yy = (*ta++) >> 8;

      yr = (yy&0x0000F8)>>3;
      yg = (yy&0x00F800)>>11;
      yb = (yy&0xF80000)>>19;

      yy = (yr<<10) | (yg<<5) | (yb<<0);

      GET_PEN_FOR_COLOUR_32(
         yr << 1,
         yg << 1,
         yb << 1,
         res
      );

      *ct++ = res;

   }while(--cols);
}

struct COLOUR_MAPPER col_map_xxxx_xxxx_rrrr_rrrr_gggg_gggg_bbbb_bbbb =
{
   "24bit xxxx xxxx rrrr rrrr gggg gggg bbbb bbbb [M68020]",
   map_24bit_xxxx_xxxx_rrrr_rrrr_gggg_gggg_bbbb_bbbb_8,
   map_24bit_xxxx_xxxx_rrrr_rrrr_gggg_gggg_bbbb_bbbb_15,
   map_24bit_xxxx_xxxx_rrrr_rrrr_gggg_gggg_bbbb_bbbb_16,
   map_24bit_xxxx_xxxx_rrrr_rrrr_gggg_gggg_bbbb_bbbb_24,
   map_24bit_xxxx_xxxx_rrrr_rrrr_gggg_gggg_bbbb_bbbb_32,
};

/******************************************************************************/
/* 24-bit xRGB (12-bit mapping) byteswapped starscream format                 */
/******************************************************************************/

void Map_24bit_xRGB_bswap(int bank, int cols)
{
   UBYTE yr,yg,yb;
   UWORD yy;
   UBYTE *ta;
   UBYTE *ct,res;

   bank_status[bank] = cols;
   ta = (RAM_PAL+(bank<<6));
   ct = coltab[bank];
   do{
      yr = (*(ta+0)) >>4;
      yg = (*(ta+3)) >>4;
      yb = (*(ta+2)) >>4;
      ta += 4;
      yy=((yr<<8)|(yg<<4)|(yb));

      GET_PEN_FOR_COLOUR_8(
         yr<<2,
         yg<<2,
         yb<<2,
         res
      );

      *ct++ = res;

   }while(--cols);
}

/******************************************************************************/
/* 21-bit xRGB (12-bit mapping) 680x0 format (Cleopatra's Fortune)            */
/******************************************************************************/

void Map_21bit_xRGB(int bank, int cols)
{
   UBYTE yr,yg,yb;
   UWORD yy;
   UBYTE *ta;
   UBYTE *ct,res;

   bank_status[bank] = cols;
   ta = (RAM_PAL+(bank<<6));
   ct = coltab[bank];
   do{
      yr = ((*(ta+1)) >>3)&0x0F;
      yg = ((*(ta+2)) >>3)&0x0F;
      yb = ((*(ta+3)) >>3)&0x0F;
      ta += 4;
      yy=((yr<<8)|(yg<<4)|(yb));

      GET_PEN_FOR_COLOUR_8(
         yr<<2,
         yg<<2,
         yb<<2,
         res
      );

      *ct++ = res;

   }while(--cols);
}

/******************************************************************************/
/* 12-bit xxxxRGBx (12-bit mapping) 680x0 format (Arabian Magic)              */
/******************************************************************************/

void Map_12bit_RGBxxxxx(int bank, int cols)
{
   UWORD yy;
   UBYTE *ta;
   UBYTE *ct,res;

   bank_status[bank] = cols;
   ta = (RAM_PAL+(bank<<6)+2);
   ct = coltab[bank];
   do{
      yy = ReadWord68k(ta)>>4;
      ta += 4;

      GET_PEN_FOR_COLOUR_8(
         (yy&0x0F00)>>6,
         (yy&0x00F0)>>2,
         (yy&0x000F)<<2,
         res
      );

      *ct++ = res;

   }while(--cols);
}

/******************************************************************************/
/* 15bit rrrr rggg ggbb bbbx xxxx xxxx xxxx xxxx                              */
/******************************************************************************/

void map_15bit_bbbb_bggg_ggrr_rrrx_xxxx_xxxx_xxxx_xxxx_8(ULONG bank, ULONG cols)
{
   UWORD yy;
   ULONG *ta;
   UBYTE *ct,res;

   bank_status[bank] = cols;
   ta = (ULONG *) (RAM_PAL+(bank<<10));
   ct = coltab[bank];
   do{
      yy = (*ta++) >> 1;

      GET_PEN_FOR_COLOUR_8(
         (yy&0x001F)<<1,
         (yy&0x03E0)>>4,
         (yy&0x7C00)>>9,
         res
      );

      *ct++ = res;

   }while(--cols);
}

void map_15bit_bbbb_bggg_ggrr_rrrx_xxxx_xxxx_xxxx_xxxx_15(ULONG bank, ULONG cols)
{
   UWORD yy;
   ULONG *ta;
   UWORD *ct,res;

   bank_status[bank] = cols;
   ta = (ULONG *) (RAM_PAL+(bank<<10));
   ct = (UWORD *) coltab[bank];
   do{
      yy = (*ta++) >> 1;

      GET_PEN_FOR_COLOUR_15(
         (yy&0x001F)<<1,
         (yy&0x03E0)>>4,
         (yy&0x7C00)>>9,
         res
      );

      *ct++ = res;

   }while(--cols);
}

struct COLOUR_MAPPER col_map_bbbb_bggg_ggrr_rrrx_xxxx_xxxx_xxxx_xxxx =
{
   "15bit bbbb bggg ggrr rrrx xxxx xxxx xxxx xxxx",
   map_15bit_bbbb_bggg_ggrr_rrrx_xxxx_xxxx_xxxx_xxxx_8,
   map_15bit_bbbb_bggg_ggrr_rrrx_xxxx_xxxx_xxxx_xxxx_15,
};

/******************************************************************************/
/* 15bit xggg ggrr rrrb bbbb                                                  */
/******************************************************************************/

void map_15bit_xggg_ggrr_rrrb_bbbb_8(ULONG bank, ULONG cols)
{
   UWORD yy;
   UWORD *ta;
   UBYTE *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) colour_ram_src[bank];
   ct = coltab[bank];
   do{
      yy = (*ta++) & 0x7FFF;

      GET_PEN_FOR_COLOUR_8(
         (yy&0x03E0)>>4,
         (yy&0x7C00)>>9,
         (yy&0x001F)<<1,
         res
      );

      *ct++ = res;

   }while(--cols);
}

void map_15bit_xggg_ggrr_rrrb_bbbb_15(ULONG bank, ULONG cols)
{
   UWORD yy;
   UWORD *ta;
   UWORD *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) colour_ram_src[bank];
   ct = (UWORD *) coltab[bank];
   do{
      yy = (*ta++) & 0x7FFF;

      GET_PEN_FOR_COLOUR_15(
         (yy&0x03E0)>>4,
         (yy&0x7C00)>>9,
         (yy&0x001F)<<1,
         res
      );

      *ct++ = res;

   }while(--cols);
}

void map_15bit_xggg_ggrr_rrrb_bbbb_16(ULONG bank, ULONG cols)
{
   UWORD yy;
   UWORD *ta;
   UWORD *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) colour_ram_src[bank];
   ct = (UWORD *) coltab[bank];
   do{
      yy = (*ta++) & 0x7FFF;

      GET_PEN_FOR_COLOUR_16(
         (yy&0x03E0)>>4,
         (yy&0x7C00)>>9,
         (yy&0x001F)<<1,
         res
      );

      *ct++ = res;

   }while(--cols);
}

void map_15bit_xggg_ggrr_rrrb_bbbb_24(ULONG bank, ULONG cols)
{
   UWORD yy;
   UWORD *ta;
   ULONG *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) colour_ram_src[bank];
   ct = (UWORD *) coltab[bank];
   do{
      yy = (*ta++) & 0x7FFF;

      GET_PEN_FOR_COLOUR_24(
         (yy&0x03E0)>>4,
         (yy&0x7C00)>>9,
         (yy&0x001F)<<1,
         res
      );

      *ct++ = res;

   }while(--cols);
}

void map_15bit_xggg_ggrr_rrrb_bbbb_32(ULONG bank, ULONG cols)
{
   UWORD yy;
   UWORD *ta;
   ULONG *ct,res;

   bank_status[bank] = cols;
   ta = (UWORD *) colour_ram_src[bank];
   ct = (UWORD *) coltab[bank];
   do{
      yy = (*ta++) & 0x7FFF;

      GET_PEN_FOR_COLOUR_32(
         (yy&0x03E0)>>4,
         (yy&0x7C00)>>9,
         (yy&0x001F)<<1,
         res
      );

      *ct++ = res;

   }while(--cols);
}

struct COLOUR_MAPPER col_map_xggg_ggrr_rrrb_bbbb =
{
   "15bit xggg ggrr rrrb bbbb",
   map_15bit_xggg_ggrr_rrrb_bbbb_8,
   map_15bit_xggg_ggrr_rrrb_bbbb_15,
   map_15bit_xggg_ggrr_rrrb_bbbb_16,
   map_15bit_xggg_ggrr_rrrb_bbbb_24,
   map_15bit_xggg_ggrr_rrrb_bbbb_32,
};

/******************************************************************************/
/*                                                                            */
/*                     CLOSEST MATCHING COLOUR ROUTINE                        */
/*                                                                            */
/******************************************************************************/

static ULONG col_diff[128*3]; 

static void closest_colour_init(void)
{
   int i;

   for(i=1; i<64; i++){
      int k = i * i;
      col_diff[0  +i] = col_diff[0  +128-i] = k * (59 * 59);
      col_diff[128+i] = col_diff[128+128-i] = k * (30 * 30);
      col_diff[256+i] = col_diff[256+128-i] = k * (11 * 11);
   }
}

/*

Find the best fitting colour in the range res_pens - 254

*/

static UBYTE closest_colour(int r, int g, int b)
{
   int i, coldiff, lowest, bestfit;
   RGB *rgb;

   bestfit = 0;
   lowest = INT_MAX;

   i = res_pens;

   rgb = &pal[i];

   do{
      coldiff = (col_diff + 0) [ (rgb->g - g) & 0x7F ];
      if (coldiff < lowest) {
	 coldiff += (col_diff + 128) [ (rgb->r - r) & 0x7F ];
	 if (coldiff < lowest) {
	    coldiff += (col_diff + 256) [ (rgb->b - b) & 0x7F ];
	    if (coldiff < lowest) {
	       bestfit = i;
	       if (coldiff == 0)
		  return bestfit;
	       lowest = coldiff;
	    }
	 }
      }
      i++;
      rgb++;
   }while(i<255);

   return bestfit;
}

// Set changed colours. Since this is a slow thing on the pc,
// I must try to reduce it to be called when only needed.

void update_screen_palette(PALETTE palette)
{
   #ifdef RDTSC_PROFILE
   if(raine_cfg.show_fps_mode>2){
      ProfileStop(PRO_BLIT);
      ProfileStart(PRO_PAL);
   }
   #endif

#if defined(RAINE_DOS) && !defined(__linux__)
   if(display_cfg.fast_set_pal)
      fast_palette_update(palette,pal_screen,display_cfg.vsync);
   else							// User has a lame gfx card/vesa driver
	  set_palette_range(palette,0,255,display_cfg.vsync);
#else
	  set_palette_range(palette,0,255,display_cfg.vsync);
#endif

   #ifdef RDTSC_PROFILE
   if(raine_cfg.show_fps_mode>2){
      ProfileStop(PRO_PAL);
      ProfileStart(PRO_BLIT);
   }
   #endif
}

static ULONG src_add,dest_add;

void raine_blit_x2(BITMAP *src, BITMAP *dest, int s_x, int s_y, int d_x, int d_y, int w, int h)
{
#ifdef RAINE_DOS

   src_add   = (src->w);
   dest_add  = (dest->line[1] - dest->line[0]) - (w+w);

   asm("
	pushal
	pushw	%%es

	movw	%%dx,%%es

	movl	%%eax,%%ebp

	movl	%%ecx,shit_03 - 4

	movl	src_add,%%eax
	movl	dest_add,%%ebx
	movl	%%eax,shit_00 - 4
	movl	%%ebx,shit_01 - 4
	.align	4
1:
	movl	$0x39393931,%%ecx
shit_03:
	pushl	%%esi
0:
	movl	(%%esi),%%eax
	movl	%%eax,%%ebx
	bswap	%%eax
	xchgw	%%ax,%%bx
	roll	$8,%%eax
	rorl	$8,%%ebx
	movl	%%eax,%%es:(%%edi)
	addl	$4,%%esi
	movl	%%ebx,%%es:4(%%edi)
	addl	$8,%%edi

	decl	%%ecx
	jnz	0b

	popl	%%esi

	testl	$0x01,%%ebp
	jz	3f
	addl	$0x39393931,%%esi
3:
shit_00:
	addl	$0x39393931,%%edi
shit_01:
	decl	%%ebp
	jne	1b

	popw	%%es
	popal
	"
	:
	: "a" (h<<1),
	  "c" (w>>2),
	  "d" (dest->seg),
	  "S" ( src->line[s_y]+s_x),
	  "D" (dest->line[d_y]+d_x)
	: "ebx"
   );
#endif
}

void raine_blit_x2_x(BITMAP *src, BITMAP *dest, int s_x, int s_y, int d_x, int d_y, int w, int h)
{
#ifdef RAINE_DOS

   src_add   = (src->w);
   dest_add  = (dest->line[1] - dest->line[0]) - (w+w);

   asm("
	pushal
	pushw	%%es

	movw	%%dx,%%es

	movl	%%eax,%%ebp

	movl	%%ecx,poop_03 - 4

	movl	src_add,%%eax
	movl	dest_add,%%ebx
	movl	%%eax,poop_00 - 4
	movl	%%ebx,poop_01 - 4
	.align	4
1:
	movl	$0x39393931,%%ecx
poop_03:
	pushl	%%esi
0:
	movl	(%%esi),%%eax
	movl	%%eax,%%ebx
	bswap	%%eax
	xchgw	%%ax,%%bx
	roll	$8,%%eax
	rorl	$8,%%ebx
	movl	%%eax,%%es:(%%edi)
	addl	$4,%%esi
	movl	%%ebx,%%es:4(%%edi)
	addl	$8,%%edi

	decl	%%ecx
	jnz	0b

	popl	%%esi

	addl	$0x39393931,%%esi
poop_00:
	addl	$0x39393931,%%edi
poop_01:
	decl	%%ebp
	jne	1b

	popw	%%es
	popal
	"
	:
	: "a" (h),
	  "c" (w>>2),
	  "d" (dest->seg),
	  "S" ( src->line[s_y]+s_x),
	  "D" (dest->line[d_y]+d_x)
	: "ebx"
   );
#endif
}

#if 0

void set_12bit_palette(void)
{
   int r, g, b;

   for (b=0; b<15; b++) {
      for (g=0; g<16; g++) {
	 pal_12bit[b*16+g].r = 0;
	 pal_12bit[b*16+g].g = g*63/15;
	 pal_12bit[b*16+g].b = b*63/15;
      }
   }

   for (r=0; r<16; r++) {
      pal_12bit[r+240].r = r*63/15;
      pal_12bit[r+240].g = 0;
      pal_12bit[r+240].b = 0;
   }
}

int makecol12(int r, int g, int b)
{
   r = r*16/256;
   g = g*16/256;
   b = b*16/256 - 1;

   if (b < 0)
      b = 0;

   return (r << 8) | (b << 4) | g | 0xF000;
}

void fake_12_bit_blit(BITMAP *src, BITMAP *dest, int s_x, int s_y, int d_x, int d_y, int w, int h)
{
   UWORD pal_map_1[256];
   UWORD pal_map_2[256];
   UBYTE *src_1;
   UWORD *dest_1;
   UWORD *dest_2;
   int tx,ty;

   set_12bit_palette();

   for(tx=0;tx<256;tx++){
      ty = makecol12(pal[tx].r << 2, pal[tx].g << 2, pal[tx].b << 2);
      pal_map_1[tx] = (((ty>>0)<<8)&0xFF00) | (((ty>>0)<<0)&0x00FF);
      pal_map_2[tx] = (((ty>>8)<<8)&0xFF00) | (((ty>>8)<<0)&0x00FF);
   }

   for(ty=0;ty<h;ty++){
   src_1  =  src->line[s_y+ty]+s_x;
   dest_1 = dest->line[d_y+ty+ty+0]+d_x;
   dest_2 = dest->line[d_y+ty+ty+1]+d_x;
   for(tx=0;tx<w;tx++){
      dest_1[tx] = pal_map_1[ src_1[tx] ];
      dest_2[tx] = pal_map_2[ src_1[tx] ];
   }
   }
}

#endif
